// FiveTech source code editor and projects manager

#include "FiveWin.ch"
#include "Splitter.ch"

#define GW_CHILD        5 
#define GW_HWNDNEXT     2
#define SW_RESTORE      9

#define PROJECT_WIDTH 190

extern TRibbonBar

static oFivedit

//----------------------------------------------------------------------------//

function Main( cFileName )

   // SetResDebug( .T. )

   if IsExeRunning( cFileName( HB_ARGV( 0 ) ) )
      ShowApplication( cFileName )
   else   		
      oFivedit = TFivedit():New( cFileName )
      
      oFivedit:Activate()
   endif   		

   // CheckRes()
   // WinExec( "notepad checkres.txt" )	

return nil

//----------------------------------------------------------------------------//

function ShowApplication( cFileName )

   local hWnd := FindWnd( cFileNoExt( "Fivedit" ) ) 
   
   if hWnd != nil
      if IsIconic( hWnd ) 
         ShowWindow( hWnd, SW_RESTORE ) 
      endif
      SetForeGroundWindow( hWnd )
      OpenClipboard( GetDesktopWindow() )
      SetClipboardData( 1, cFileName )
      CloseClipboard()
      SendMessage( hWnd, FM_OPENFILE )
   endif   
   
return nil   

//----------------------------------------------------------------------------//

function FindWnd( cTitle ) 

   local hWnd := GetWindow( GetDesktopWindow(), GW_CHILD ) 

   while hWnd != 0 
      if Upper( cTitle ) $ Upper( GetWindowText( hWnd ) ) 
         return hWnd
      endif 

      hWnd = GetWindow( hWnd, GW_HWNDNEXT ) 
   end 

return nil

//----------------------------------------------------------------------------//

CLASS TFivedit

   DATA oWnd, oFld, oBar, oEditor, oRowItem, oColItem, oMsgBar
   DATA oVSplitL, oHSplitC, oVSplitR, oBtnSave
   DATA nLeftPos INIT PROJECT_WIDTH
   DATA nMiddlePos, nRightPos
   DATA oFunList, aFunLines, oItemFuncList
   DATA oTree, oItemProject
   DATA oMemo, cOutPut, oItemResult
   DATA cParameters, cDirName INIT GetCurDir()
   DATA oMruPrgs, oMruPrjs, oMenuTools

   DATA cFWHPath INIT PadR( "c:\fwh", 200 )
   DATA cFWHMode INIT "32 bits"
   DATA cFWHExtraLibs INIT Space( 200 )

   DATA cHarbPath INIT PadR( "c:\harbour", 200 )
   DATA cHarbComp INIT "Harbour " 
   DATA cHarbFlags INIT PadR( "/n /w", 100 )
   DATA cHarbExtraLibs INIT Space( 200 )

   DATA cCCompilerPath INIT PadR( "c:\bcc582", 200 )
   DATA cCCompilerFlags INIT PadR( "-Gn -aa -Tpe -s", 100 )
   DATA cCCompilerExtraLibs INIT Space( 200 )
   
   DATA nTextColor INIT 0
   DATA nBackColor INIT nRGB( 160, 160, 160 )
   DATA nCaretBackColor INIT nRGB( 0, 179, 179 )
   
   DATA cFontInfo INIT ;
      PadR( '{"Courier New",0,-12,.F.,.F.,0,0,,.F.,.F.,.F.,0,3,2,1,,49}', 200 )
   
   DATA cTheme INIT FWString( "Standard" )
   	
   DATA lDuplicates INIT .T.
   
   DATA nLanguage INIT 1

   METHOD New( cFileName ) CONSTRUCTOR

   METHOD Activate()

   METHOD AddBorlandLibs()
   
   METHOD AddHarbourLibs()

   METHOD BuildButtonBar()
   
   METHOD BuildEditor( cFileName )
   
   METHOD BuildLeft()
   
   METHOD BuildMenu()
   
   METHOD BuildMsgBar()
   
   METHOD BuildProject( lDebug )
   
   METHOD BuildRight()
   
   METHOD CheckBorlandSetupFiles()   
   
   METHOD Close()
   
   METHOD CloseProject()
   
   METHOD CompilePRG()
   
   METHOD CompileRC( cFileName )
   
   METHOD CopyFuncListToClipboard()	
   
   METHOD Exit()
   
   METHOD FillFuncList()
   
   METHOD HideFuncList()
   	
   METHOD HideProject()	
   
   METHOD HideResult()	
   
   METHOD Link()
   
   METHOD LoadPanelsSettings()
   	
   METHOD LoadPreferences( cIniFile )
   
   METHOD MenuSetTools()
   
   METHOD NewFile()
   
   METHOD OpenFile( cFileName )
   
   METHOD OpenProject( cPrjName )
   
   METHOD Parameters()
   
   METHOD Preferences()
   
   METHOD ProjectAddItem()
   
   METHOD ProjectDelItem()
   
   METHOD ProjectOpenFile()
   
   METHOD Run( lDebug )
   
   METHOD RunScript()
   
   METHOD SavePanelsSettings()	
   
   METHOD SavePreferences( cIniFile )
   
   METHOD SaveProject()
   
   METHOD SearchFileInFolder( cFileName, lShow )
   
   METHOD SearchFileInTree( cFileName )
   
   METHOD SelectFunction()
   
   METHOD SetDarkTheme( oGetTextColor, oGetBackColor, oGetCaretBackColor )	
   
   METHOD ShowFuncList()
   	
   METHOD ShowProject()	
   
   METHOD ShowRightPopup( nRow, nCol, oWnd )	
   
   METHOD ShowResult()	
   
   METHOD Tools()

ENDCLASS

//----------------------------------------------------------------------------//

METHOD New( cFileName ) CLASS TFivedit

   ::nRightPos  = ScreenWidth() - 200
   ::nMiddlePos = ScreenHeight() - 200 

   SetDlgGradient( { { 1, RGB( 199, 216, 237 ), RGB( 237, 242, 248 ) } } )
   
   ::LoadPreferences()
   
   // DEFINE WINDOW ::oWnd FROM 1, 1 TO 400, 600 PIXEL ;
   //    TITLE "Fivedit" MENU ::BuildMenu()
   
   ::oWnd := TFiveditWnd():New( 1, 1, 400, 600, "Fivedit",, ::BuildMenu(),;
                                ,,,,,,,,, .T., .T., .T., .T., .T.,, "::oWnd", )
         
   ::oWnd:Maximize()   
      
   ::BuildButtonBar() 
   ::BuildLeft()  
   ::BuildRight()
   ::BuildMsgBar()

   if ! Empty( cFileName ) .and. File( cFileName )
      ::OpenFile( cFileName )
   else
      ::BuildEditor()
   endif    

   @ ::nMiddlePos + 4, ::nLeftPos + 1 GET ::oMemo VAR ::cOutPut MEMO READONLY ;
      SIZE ::nRightPos - ::nLeftPos, ::oMsgBar:nTop - ::nMiddlePos - 4 ;
      OF ::oWnd PIXEL 

   ::oMemo:SetColor( 0, CLR_GRAY )

   @ ::nMiddlePos, ::nLeftPos + 1 SPLITTER ::oHSplitC ;
              HORIZONTAL ;
              PREVIOUS CONTROLS ::oFld ;
              HINDS CONTROLS ::oMemo ;
              TOP MARGIN 80 ;
              BOTTOM MARGIN 80 ;
              SIZE ::nRightPos - ::nLeftPos - 2, 4 PIXEL ;
              OF ::oWnd ;
              STYLE

   @  0, ::nLeftPos - 5 SPLITTER ::oVSplitL ;
              VERTICAL ;
              PREVIOUS CONTROLS ::oTree ; 
              HINDS CONTROLS ::oFld, ::oHSplitC, ::oMemo ;
              LEFT MARGIN 0 ;
              RIGHT MARGIN 80 ;
              SIZE 4, ScreenHeight() - ::oBar:nHeight - ;
              ::oMsgBar:nHeight PIXEL ;
              OF ::oWnd ;
              STYLE

   @  61, ::nRightPos SPLITTER ::oVSplitR ;
             VERTICAL ;
             PREVIOUS CONTROLS ::oFld, ::oHSplitC, ::oMemo ;
             HINDS CONTROLS ::oFunList ;
             LEFT MARGIN 80 ;
             RIGHT MARGIN 80 ;
             SIZE 4, ScreenHeight() - ::oBar:nHeight - ;
             ::oMsgBar:nHeight - 19 PIXEL ;
             OF ::oWnd ;
             STYLE

   ::oTree:Refresh()

return Self

//----------------------------------------------------------------------------//

METHOD Activate() CLASS TFivedit

   local oThis := Self

   ACTIVATE WINDOW ::oWnd MAXIMIZED ;
      ON INIT   oThis:LoadPanelsSettings() ;	
      ON RESIZE If( nSizeType != 1, ( ::oVSplitL:AdjLeft(),;
                  If( ::oItemFuncList:lChecked,;
                  ( ::oVSplitR:SetPosition( ::oWnd:nWidth - 200 ),;
                    ::oVSplitR:Adjust(), ::oFunList:Refresh() ),),;
                  If( ::oItemResult:lChecked,;
                  ::oHSplitC:SetPosition( ::oWnd:nHeight - 200 ),),;
                  ::oEditor:SetFocus() ), )
                  
   ::SavePanelsSettings()               						

return nil

//----------------------------------------------------------------------------//

METHOD AddBorlandLibs() CLASS TFivedit

   local cLibs, cExtraLibs := ""

   if ! Empty( ::cCCompilerExtraLibs )
      AEval( hb_aTokens( StrTran( AllTrim( ::cCCompilerExtraLibs ), ",", "" ) ),;
             { | cLib | cExtraLibs += AllTrim( ::cCCompilerPath ) + "\lib\" + ;
                        cLib + " " } )
   endif                     

   cLibs := AllTrim( ::cCCompilerPath ) + "\lib\" + "cw32.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\" + "uuid.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\" + "import32.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\psdk\" + "odbc32.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\psdk\" + "nddeapi.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\psdk\" + "msimg32.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\psdk\" + "psapi.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\psdk\" + "rasapi32.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\psdk\" + "gdiplus.lib " + ;
            AllTrim( ::cCCompilerPath ) + "\lib\psdk\" + "shell32.lib " + ;
            cExtraLibs

return cLibs

//----------------------------------------------------------------------------//

METHOD AddHarbourLibs() CLASS TFivedit

   local lHarbour := AllTrim( ::cHarbComp ) == "Harbour"
   local cLibs, cFWHExtraLibs := "", cHarbExtraLibs := "" 

   if ! Empty( ::cFWHExtraLibs )
      AEval( hb_aTokens( StrTran( AllTrim( ::cFWHExtraLibs ), ",", "" ) ),;
             { | cLib | cFWHExtraLibs += AllTrim( ::cFWHPath ) + "\lib\" + ;
                        cLib + " " } )
   endif                     

   if ! Empty( ::cHarbExtraLibs )
      AEval( hb_aTokens( StrTran( AllTrim( ::cHarbExtraLibs ), ",", "" ) ),;
             { | cLib | cHarbExtraLibs += AllTrim( ::cHarbPath ) + "\lib\" + ;
                        cLib + " " } )
   endif                     

   cLibs = AllTrim( ::cFWHPath ) + "\lib\" + ;
           If( lHarbour, "fiveh.lib ", "fivehx.lib " ) + ;
           AllTrim( ::cFWHPath ) + "\lib\" + "fivehc.lib " + ;
           cFWHExtraLibs + ;
           If( lHarbour, AllTrim( ::cHarbPath ) + "\lib\" + "hbwin.lib ", "" ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + "gtgui.lib " + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbrtl.lib ", "rtl.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbvm.lib ", "vm.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hblang.lib ", "lang.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbmacro.lib ", "macro.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbrdd.lib ", "rdd.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "rddntx.lib ", "dbfntx.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "rddcdx.lib ", "dbfcdx.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "rddfpt.lib ", "dbffpt.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + "hbsix.lib " + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbdebug.lib ", "debug.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbcommon.lib ", "common.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbpp.lib ", "pp.lib " ) + ;
           If( lHarbour, AllTrim( ::cHarbPath ) + "\lib\" + ;
           "hbcpage.lib ", "" ) + ;
           If( lHarbour, AllTrim( ::cHarbPath ) + "\lib\" + ;
           "hbcplr.lib ", "" ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbct.lib ", "ct.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbpcre.lib ", "pcrepos.lib " ) + ;
           If( lHarbour, AllTrim( ::cHarbPath ) + "\lib\" + ;
           "xhb.lib ", "" ) + ;
           If( lHarbour, AllTrim( ::cHarbPath ) + "\lib\" + ;
           "hbziparc.lib ", "" ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbmzip.lib ", "hbzip.lib " ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + ;
           If( lHarbour, "hbzlib.lib ", "zlib.lib " ) + ;
           If( lHarbour, AllTrim( ::cHarbPath ) + "\lib\" + ;
           "minizip.lib ", "" ) + ;
           AllTrim( ::cHarbPath ) + "\lib\" + "png.lib " + ;
           If( lHarbour .and. File( AllTrim( ::cHarbPath ) + "\lib\" + "hbcurl.lib" ),;
           		AllTrim( ::cHarbPath ) + "\lib\" + "hbcurl.lib ", "" ) + ;
           If( lHarbour .and. File( AllTrim( ::cHarbPath ) + "\lib\" + "hbusrrdd.lib" ),;
               AllTrim( ::cHarbPath ) + "\lib\" + "hbusrrdd.lib ", "" ) + ;
           cHarbExtraLibs			  

return cLibs

//----------------------------------------------------------------------------//

METHOD BuildButtonBar() CLASS TFivedit

   DEFINE BUTTONBAR ::oBar SIZE 60, 60 OF ::oWnd 2007

   DEFINE BUTTON OF ::oBar RESOURCE "new" TOOLTIP FWString( "New" ) ;
      PROMPT FWString( "New" ) ACTION ::NewFile()

   DEFINE BUTTON OF ::oBar RESOURCE "open" ;
      PROMPT FWString( "Open" ) ACTION ::OpenFile() ;
      TOOLTIP FWString( "Open" )

   DEFINE BUTTON OF ::oBar RESOURCE "close" ;
      PROMPT FWString( "Close" ) ACTION ::Close() ;
      TOOLTIP FWString( "Close" )

   DEFINE BUTTON ::oBtnSave OF ::oBar RESOURCE "floppy" ;
      PROMPT FWString( "Save" ) ;
      ACTION ( ::oEditor:Save(), ::oBtnSave:Disable() ) ;
      TOOLTIP FWString( "Save" )

   ::oBtnSave:Disable()

   DEFINE BUTTON OF ::oBar GROUP RESOURCE "play" ;
      PROMPT "Script" ACTION ::RunScript() ;
      TOOLTIP FWString( "Run as script" )

   DEFINE BUTTON OF ::oBar RESOURCE "run" ;
      PROMPT FWString( "Run" ) ;
      ACTION ( ::Run(), ::oEditor:SetFocus() ) ;
      TOOLTIP FWString( "Build and execute" )

   DEFINE BUTTON OF ::oBar RESOURCE "debug" ;
      PROMPT FWString( "Debug" ) ;
      ACTION ::Run( .T. ) ;
      TOOLTIP FWString( "Build and execute with debugger" )

   DEFINE BUTTON OF ::oBar GROUP RESOURCE "undo" ;
      PROMPT FWString( "Undo" ) ;
      ACTION ( ::oEditor:UnDo(), ::oEditor:SetFocus() ) ;
      TOOLTIP FWString( "UnDo" )

   DEFINE BUTTON OF ::oBar RESOURCE "redo" ;
      PROMPT FWString( "Redo" ) ;
      ACTION ( ::oEditor:ReDo(), ::oEditor:SetFocus() ) ;
      TOOLTIP FWString( "ReDo" )

   DEFINE BUTTON OF ::oBar GROUP RESOURCE "cut" ;
      PROMPT FWString( "Cut" ) ACTION ::oEditor:Cut() ;
      TOOLTIP FWString( "Cut" )

   DEFINE BUTTON OF ::oBar RESOURCE "copy" ;
      PROMPT FWString( "Copy" ) ACTION ::oEditor:Copy() ;
      TOOLTIP FWString( "Copy" )

   DEFINE BUTTON OF ::oBar RESOURCE "paste" ;
      PROMPT FWString( "Paste" ) ACTION ::oEditor:Paste() ;
      TOOLTIP FWString( "Paste" )

   DEFINE BUTTON OF ::oBar GROUP RESOURCE "search" ;
      PROMPT FWString( "Search" ) ACTION ::oEditor:DlgFindText() ;
      TOOLTIP FWString( "Search" )

   DEFINE BUTTON OF ::oBar RESOURCE "previous" ;
      PROMPT FWString( "Previous" ) ACTION ::oEditor:FindPrev() ;
      TOOLTIP FWString( "Find previous" )

   DEFINE BUTTON OF ::oBar RESOURCE "next" ;
      PROMPT FWString( "Next" ) ACTION ::oEditor:FindNext() ;
      TOOLTIP FWString( "Find next" )

   DEFINE BUTTON OF ::oBar RESOURCE "replace" ;
      PROMPT FWString( "Replace" ) ACTION ::oEditor:DlgReplace() ;
      TOOLTIP FWString( "Replace" )

   DEFINE BUTTON OF ::oBar RESOURCE "gotoline" ;
      PROMPT FWString( "GotoLine" ) ;
      ACTION ( ::oEditor:DlgGotoLine(), ::oEditor:SetFocus() ) ;
      TOOLTIP FWString( "Goto Line" )

   DEFINE BUTTON OF ::oBar GROUP RESOURCE "help" ;
      PROMPT FWString( "Help" ) ACTION HelpIndex() ;
      TOOLTIP FWString( "Help" )

   DEFINE BUTTON OF ::oBar RESOURCE "exit" ;
      PROMPT FWString( "Exit" ) ACTION ::Exit() ;
      TOOLTIP FWString( "Exit" )

return nil

//----------------------------------------------------------------------------//

METHOD BuildEditor( cFileName ) CLASS TFivedit

   DEFAULT cFileName := "noname.prg"

   if ::oFld == nil
      @ ::oBar:nHeight - 1, ::nLeftPos FOLDER ::oFld PROMPTS "" OF ::oWnd ;
         SIZE ::nRightPos - ::nLeftPos, ::nMiddlePos - ::oBar:nHeight + 2 PIXEL 
      ::oFld:DelPages()   	
   endif
   
   ::oFld:AddItem( cFileNoPath( cFileName ) )
   ::oFld:bResized = { || ::oEditor:AdjClient() }
   ::oEditor = TScintilla():New( 0, 0, ;
                                 ::nRightPos - ::nLeftPos, ::nMiddlePos - ;
                                 ::oBar:nHeight + 2,;
                                 ATail( ::oFld:aDialogs ) )
                
   ::oEditor:cFileName = cFileName                            
   ::oEditor:Setup()

   ::oEditor:SetColor( ::nTextColor, ::nBackColor )
   
   ::oEditor:SetCaretLineBack( ::nCaretBackColor )
   
   if ! Empty( ::cFontInfo )
      ::oEditor:SetFont( FontFromText( ::cFontInfo ) )
   endif   
   
   ::oEditor:oRowItem = ::oRowItem
   ::oEditor:oColItem = ::oColItem
   ::oEditor:bChange = { || ::SelectFunction(), If( ::oEditor:GetModify(),;
                            ::oBtnSave:Enable(), ::oBtnSave:Disable() ) }
   ::oEditor:bGotFocus = { | o | ::oEditor := o, ::FillFuncList() } 
   ::oEditor:bKeyDown = { | nKey | If( nKey == VK_RETURN,;
                                       ::FillFuncList(),) ;
                                   /* If( nKey == VK_RETURN,;
                                       ::oEditor:AutoIndent(),) */ }
   ::oEditor:cFileName = cFileName
   
   if File( cFileName )
      ::oEditor:SetText( MemoRead( cFileName ) )
   else   
      ::oEditor:SetText( '#include "FiveWin.ch"' + CRLF + CRLF + ;
                         "function Main()" + CRLF + CRLF + ;
                         '   MsgInfo( "Hello world!" )' + CRLF + CRLF + ;
                         "return nil" )
   endif                      

   ::oFld:bChange = { | nOption | ::SearchFileInTree( ::oFld:aPrompts[ nOption ] ) }

   ::oEditor:AdjClient()  
   
   ::FillFuncList()               

return nil

//----------------------------------------------------------------------------//

METHOD SavePanelsSettings() CLASS TFivedit

   local oIni
  
   lMkDir( "c:\users\" + WNetGetUser() + "\appdata\local\fivedit" )

   INI oIni FILENAME "c:\users\" + WNetGetUser() + ;
     	               "\appdata\local\fivedit\fivedit.ini"
      SET SECTION "panels" ENTRY "1" ;
         TO If( ::oItemProject:lChecked, ".T.", ".F." ) OF oIni

      SET SECTION "panels" ENTRY "2" ;
         TO If( ::oItemFuncList:lChecked, ".T.", ".F." ) OF oIni

      SET SECTION "panels" ENTRY "3" ;
         TO If( ::oItemResult:lChecked, ".T.", ".F." ) OF oIni
   ENDINI      

return nil

//----------------------------------------------------------------------------//

METHOD SavePreferences( cIniFile ) CLASS TFivedit

   local oIni, n

   DEFAULT cIniFile := "c:\users\" + WNetGetUser() + ;
       	  "\appdata\local\fivedit\fivedit.ini"

   INI oIni FILENAME cIniFile
      SET SECTION "preferences" ENTRY "1" ;
         TO AllTrim( ::cFWHPath ) OF oIni
         
      SET SECTION "preferences" ENTRY "2" ;
         TO AllTrim( ::cFWHMode ) OF oIni
         
      SET SECTION "preferences" ENTRY "3" ;
         TO AllTrim( ::cFWHExtraLibs ) OF oIni
      
      SET SECTION "preferences" ENTRY "4" ;
         TO AllTrim( ::cHarbPath ) OF oIni

      SET SECTION "preferences" ENTRY "5" ;
         TO AllTrim( ::cHarbComp ) OF oIni
         
      SET SECTION "preferences" ENTRY "6" ;
         TO AllTrim( ::cHarbFlags ) OF oIni
         
      SET SECTION "preferences" ENTRY "7" ;
         TO AllTrim( ::cHarbExtraLibs ) OF oIni
      
      SET SECTION "preferences" ENTRY "8" ;
         TO AllTrim( ::cCCompilerPath ) OF oIni
         
      SET SECTION "preferences" ENTRY "9" ;
         TO AllTrim( ::cCCompilerFlags ) OF oIni
         
      SET SECTION "preferences" ENTRY "10" ;
         TO AllTrim( ::cCCompilerExtraLibs ) OF oIni

      SET SECTION "preferences" ENTRY "11" TO ::nTextColor OF oIni
         
      SET SECTION "preferences" ENTRY "12" TO ::nBackColor OF oIni

      SET SECTION "preferences" ENTRY "13" TO ::nCaretBackColor OF oIni
         
      SET SECTION "preferences" ENTRY "14" ;
         TO AllTrim( ::cFontInfo ) OF oIni

      SET SECTION "preferences" ENTRY "15" ;
         TO AllTrim( ::cTheme ) OF oIni

      SET SECTION "preferences" ENTRY "16" TO ::nLanguage OF oIni

      SET SECTION "preferences" ENTRY "17" TO ::lDuplicates OF oIni
      
      SET SECTION "preferences" ENTRY "18" TO ::cDirName OF oIni
   ENDINI

return nil

//----------------------------------------------------------------------------//

METHOD SaveProject() CLASS TFivedit

   local cProject := "[files]" + CRLF
   local n
   
   for n = 1 to Len( ::oTree:aItems[ 1 ]:aItems )
      cProject += AllTrim( Str( n ) ) + "=" + ;
                  ::oTree:aItems[ 1 ]:aItems[ n ]:cPrompt + CRLF
   next
   
   if Empty( ::oTree:Cargo )
      ::oTree:Cargo = cGetFile( "Project (*.prj) | *.prj", "Save as", 0,;
                                hb_CurDrive() + ":\" + CurDir(), .T., .T. )
   endif                           
   
   if ! Empty( ::oTree:Cargo )
      ::oTree:aItems[ 1 ]:SetText( ::oTree:Cargo )
      ::oMruPrjs:Save( ::oTree:Cargo )
      MemoWrit( ::oTree:Cargo, cProject )  
      ::SavePreferences( ::oTree:Cargo )             
   endif   

return nil

//----------------------------------------------------------------------------//

METHOD SearchFileInFolder( cFileName, lShow ) CLASS TFivedit

   local nAt := AScan( ::oFld:aDialogs,;
                       { | oDlg | If( Len( oDlg:aControls ) > 0,;
                       oDlg:aControls[ 1 ]:cFileName == cFileName, .F. ) } )
   
   DEFAULT lShow := .T.
   
   if nAt != 0 .and. lShow
      ::oFld:SetOption( nAt )
      ::oEditor = ::oFld:aDialogs[ ::oFld:nOption ]:aControls[ 1 ]
      ::oEditor:SetFocus()
      ::FillFuncList() 
   endif   

return nAt

//----------------------------------------------------------------------------//

METHOD SearchFileInTree( cFileName ) CLASS TFivedit

   local oItem := ::oTree:Scan( { | o | cFileName $ o:cPrompt } )
   
   if oItem != nil
      ::oTree:Select( oItem )
      return .T.
   endif   
   
return .F.

//----------------------------------------------------------------------------//

METHOD BuildLeft() CLASS TFivedit

   local oImageList, oBmp1, oBmp2

   oImageList = TImageList():New()

   oImageList:Add( oBmp1 := TBitmap():Define( "folder",,  ::oWnd ),;
                   oBmp2 := TBitmap():Define( "fldmask",, ::oWnd ) )
   oBmp1:End()
   oBmp2:End()	
   oImageList:Add( oBmp1 := TBitmap():Define( "source",,  ::oWnd ),;
                   oBmp2 := TBitmap():Define( "srcmask",, ::oWnd ) )
   oBmp1:End()
   oBmp2:End()	
   oImageList:Add( oBmp1 := TBitmap():Define( "icon",,    ::oWnd ),;
                   oBmp2 := TBitmap():Define( "icomask",, ::oWnd ) )
   oBmp1:End()
   oBmp2:End()	
   oImageList:Add( oBmp1 := TBitmap():Define( "bitmap",,  ::oWnd ),;
                   oBmp2 := TBitmap():Define( "bmpmask",, ::oWnd ) )
   oBmp1:End()
   oBmp2:End()	

   ::oTree = TTreeView():New( 3, 0, ::oWnd )
   ::oTree:nWidth = PROJECT_WIDTH - 5
   ::oTree:SetImageList( oImageList )
   ::oTree:Add( "Project" )
   ::oTree:Expand()
   ::oTree:SetColor( 0, CLR_GRAY )
   
   ::oTree:bChanged = { | oTree, oItem | oItem := oTree:GetSelected(),;
      If( oItem != nil, ::SearchFileInFolder( oTree:GetSelected():cPrompt ),) }
   
   ::oTree:bLDblClick = { || ::ProjectOpenFile() }

return nil

//----------------------------------------------------------------------------//

METHOD BuildMenu() CLASS TFivedit

   local oMenu 
   
   MENU oMenu 2007
      MENUITEM FWString( "&File" )
      MENU
         MENUITEM FWString( "&New" ) ACTION ::NewFile() ;
            RESOURCE "new16" ;
            MESSAGE FWString( "Create a new file in a new edit window" ) 
            
         MENUITEM FWString( "&Open" ) + "..." + Chr( 9 ) + "Ctrl+O" ;
            ACTION ( ::OpenFile(), ::oEditor:SetFocus() ) ;
            RESOURCE "open16" ;
            MESSAGE FWString( "Locate and open a file" ) ;
            ACCELERATOR ACC_CONTROL, Asc( "O" )

         MENUITEM FWString( "&Save" ) + Chr( 9 ) + "Ctrl+S" ;
            RESOURCE "save16" ;
            MESSAGE FWString( "Save the current file" ) ;
            ACCELERATOR ACC_CONTROL, Asc( "S" ) ;
            ACTION ::oEditor:Save()

         MENUITEM FWString( "Save &as" ) + "..." ;
            ACTION ( ::oEditor:cFileName := ;
                     cGetFile( "Program (*.prg) | *.prg", "Save as", 0,;
                     hb_CurDrive() + ":\" + CurDir(), .T., .T. ),;
                     ::oEditor:Save() )

         MENUITEM FWString( "Close" ) ;
            RESOURCE "close16" ;
            MESSAGE FWString( "Close the current file" ) ;
            ACTION ::Close()
            
         SEPARATOR

         MENUITEM FWString( "&Print" ) ;
            RESOURCE "print16" ;
            MESSAGE FWString( "Print the current file" )

         MENUITEM FWString( "P&rinter setup" ) + "..." ACTION PrinterSetup()
         
         SEPARATOR
         
         MENUITEM FWString( "&Change Directory" ) + "..." ;
            ACTION If( ! Empty( ::cDirName := cGetDir( GetCurDir() ) ),;
                       ( SetCurDir( ::cDirName ), ::SavePreferences() ),)
         
         MENUITEM FWString( "E&xit" ) + Chr( 9 ) + "Alt+F4" ;
            ACTION ::oWnd:End() ;
            RESOURCE "exit16" ;
            MESSAGE FWString( "Exit this app" )

         lMkDir( "c:\users\" + WNetGetUser() + ;
          		  "\appdata\local\fivedit\fivedit.ini" )

         MRU ::oMruPRGs ;
            FILENAME "c:\users\" + WNetGetUser() + ;
             			"\appdata\local\fivedit\fivedit.ini" ;      
            SECTION  "Recent-PRG-files" ;   
            ACTION   ::OpenFile( cMruItem ) ; // automatically provided
            MESSAGE  FWString( "Open this file" ) ; 
            SIZE     10 
      ENDMENU

      MENUITEM FWString( "&Project" )
      MENU
         MENUITEM FWString( "&Open project" ) + "..." ;
            RESOURCE "open16" ACTION ::OpenProject() 
              
         MENUITEM FWString( "&Save project" ) + "..." ;
            RESOURCE "save16" ACTION ::SaveProject()
         
         MENUITEM FWString( "&Close project" ) ;
            RESOURCE "close16" ACTION ::CloseProject()
            
         SEPARATOR   
         
         MENUITEM FWString( "&Add item" ) + "..." ;
            RESOURCE "plus" ;
            ACTION ::ProjectAddItem()
               
         MENUITEM FWString( "&Delete item" ) ;
            RESOURCE "minus" ;
            ACTION ::ProjectDelItem()

         MRU ::oMruPRJs ;
            FILENAME "c:\users\" + WNetGetUser() + ;
             			"appdata\local\fivedit\fivedit.ini" ;          
            SECTION  "Recent-Project-files" ;   
            ACTION   ::OpenProject( cMruItem ) ; // automatically provided
            MESSAGE  FWString( "Open this project" ) ;
            SIZE     10 
      ENDMENU
      
      MENUITEM FWString( "&Edit" )
      MENU
         MENUITEM FWString( "&Undo" ) + Chr( 9 ) + "Alt+BkSp" ;       
            RESOURCE "undo16" ;
            MESSAGE FWString( "Undo the previous editor action" ) ;
            ACTION ::oEditor:Undo()

         MENUITEM FWString( "&Redo" ) + Chr( 9 ) + "Alt+Shift+BkSp" ; 
            RESOURCE "redo16" ;
            MESSAGE FWString( "Redo the previous undo editor action" ) ;
            ACTION ::oEditor:Redo()

         SEPARATOR

         MENUITEM FWString( "Cu&t" ) + Chr( 9 ) + "Shift+Del" ;       
            RESOURCE "cut16" ;
            MESSAGE FWString( "Remove the selected text and put it on the clipboard" ) ;
            ACTION ::oEditor:Cut()

         MENUITEM FWString( "&Copy" ) + Chr( 9 ) + "Ctrl+Ins" ;      
            RESOURCE "copy16" ;
            MESSAGE FWString( "Copy the selected text to the clipboard" ) ;
            ACTION ::oEditor:Copy()

         MENUITEM FWString( "&Paste" ) + Chr( 9 ) + "Shift+Ins" ;    
            RESOURCE "paste16" ;
            MESSAGE FWString( "Insert text from the clipboard at the current position" ) ;
            ACTION ::oEditor:Paste()

         MENUITEM FWString( "Select &All" ) ;
            MESSAGE FWString( "Select all text in the editor" ) ;
            ACTION ::oEditor:SelectAll()

         SEPARATOR

         MENUITEM FWString( "Line &Duplicate" ) + Chr( 9 ) + "F5" ;
            RESOURCE "dupline16" ;
            MESSAGE FWString( "Duplicate line actual" ) ;
            ACTION ::oEditor:LineDuplicate() ;
            ACCELERATOR 0, VK_F5

         MENUITEM FWString( "Code &separator" ) + Chr( 9 ) + "F4" ;
            RESOURCE "dupline16" ;
            MESSAGE FWString( "Insert a code separator" ) ;
            ACTION ::oEditor:InsertText( ::oEditor:GetCurrentPos(),;
                 	 "//" + Replicate( "-", 76 ) + "//" + hb_eol() ) ;
            ACCELERATOR 0, VK_F4

      ENDMENU

      MENUITEM FWString( "&Search" )
      MENU
         MENUITEM FWString( "&Find" ) + "..." ACTION ::oEditor:DlgFindText() ;       
            RESOURCE "find16" ;
            MESSAGE FWString( "Search for text" )

         MENUITEM FWString( "Find &Next" ) + Chr( 9 ) + "F3" ;
            RESOURCE "next16" ;
            MESSAGE FWString( "Find next occurrence" ) ;
            ACTION ::oEditor:FindNext()

         MENUITEM FWString( "Find &Prev" ) + Chr( 9 ) + "Ctrl+F3" ;
            RESOURCE "previous16" ;
            MESSAGE FWString( "Find previous occurrence" ) ;
            ACTION ::oEditor:FindPrev()

         MENUITEM FWString( "&Replace" ) + "..." ;
            RESOURCE "replace16" ;
            ACTION ::oEditor:DlgReplace()            
               
         SEPARATOR
         
         MENUITEM FWString( "&Go to line number" ) + "..." ;
            ACTION ::oEditor:DlgGoToLine() ;
            RESOURCE "goto16" ;
            MESSAGE FWString( "Move cursor to a specific line number" ) 
              
      ENDMENU

      MENUITEM FWString( "&Run" )
      MENU
         MENUITEM FWString( "Run as &script" ) + Chr( 9 ) + "Ctrl+F9" ;
            RESOURCE "script16" ;
            ACTION ::RunScript() ;
            MESSAGE FWString( "Run the current program as a script without building an EXE" )
            
         MENUITEM FWString( "&Run as .EXE" ) + Chr( 9 ) + "Shift+F9" ;
            ACTION ::Run() ;
            RESOURCE "run16" ;
            MESSAGE FWString( "Build and run as an EXE" )

         SEPARATOR

         MENUITEM FWString( "De&bug" ) ;
            RESOURCE "debug16" ;
            MESSAGE FWString( "Debug the EXE" ) ;
            ACTION ::Run( .T. )

         SEPARATOR 	     
         
         MENUITEM FWString( "&Arguments" ) + "..." ACTION ::Parameters()      
      ENDMENU

      MENUITEM FWString( "&Tools" )
      MENU ::oMenuTools
         ::MenuSetTools()
      ENDMENU

      MENUITEM FWString( "&View" )
      MENU
         MENUITEM ::oItemProject PROMPT FWString( "&Project panel" ) CHECKED ;
            ACTION ( If( ::oItemProject:lChecked, ::HideProject(), ::ShowProject() ),;
                     ::oItemProject:SetCheck( ! ::oItemProject:lChecked ) ) 
               
         MENUITEM ::oItemResult PROMPT FWString( "&Result panel" ) CHECKED ;
            ACTION ( If( ::oItemResult:lChecked, ::HideResult(), ::ShowResult() ),;
                     ::oItemResult:SetCheck( ! ::oItemResult:lChecked ) )
             
         MENUITEM ::oItemFuncList PROMPT FWString( "&Functions panel" ) CHECKED ;
            ACTION ( If( ::oItemFuncList:lChecked, ::HideFuncList(), ::ShowFuncList() ),;
                     ::oItemFuncList:SetCheck( ! ::oItemFuncList:lChecked ) )  
         
         SEPARATOR	      
         
         MENUITEM FWString( "&Preferences" ) + "..." ;
            RESOURCE "setup16" ;
            ACTION ::Preferences()    
      ENDMENU

      oMenu:AddHelp( FWString( "FiveTech source code editor" ),;
                     "(c) FiveTech Software" )
   ENDMENU
   
return oMenu

//----------------------------------------------------------------------------//   

METHOD BuildMsgBar() CLASS TFivedit

   DEFINE MSGBAR ::oMsgBar OF ::oWnd ;
      PROMPT FWString( "Source code editor and projects manager" ) 2007  
   
   DEFINE MSGITEM ::oRowItem PROMPT FWString( "Row" ) + ": 0" ;
      OF ::oMsgBar SIZE 60 ;
      ACTION ::oEditor:DlgGoToLine()
   
   DEFINE MSGITEM ::oColItem PROMPT FWString( "Col" ) + ": 0" ;
      OF ::oMsgBar SIZE 60
   
   ::oMsgBar:KeybOn()
   
   ::oMsgBar:oKeyIns:lActive = .T.
   ::oMsgBar:oKeyIns:bMsg = nil

return nil

//----------------------------------------------------------------------------//

METHOD BuildProject( lDebug ) CLASS TFivedit

   local n, cFileName, cTargetFileName, oEditor
   local cEXEName, cPrjPath, nAt

   if Empty( ::oTree:Cargo )
      ::SaveProject()
   endif
   
   cPrjPath = cFilePath( ::oTree:Cargo )   

   if ! ::oItemResult:lChecked
      ::oItemResult:SetCheck( .T. )
      ::ShowResult()
   endif   		

   for n = 1 to Len( ::oTree:aItems[ 1 ]:aItems )
      cFileName = ::oTree:aItems[ 1 ]:aItems[ n ]:cPrompt 
      
      if ! File( cFileName )
         MsgAlert( FWString( "Project filename" ) + " " + ;
                   cFileName + " " + FWString( "is missing" ) + CRLF + ;
                   FWString( "Can't build the project" ) )
         return nil
      else
         if n == 1
            ::oMemo:SetText( "" )
            FErase( cPrjPath + "build.log" )
            cEXEName = cPrjPath + cFileNoExt( cFileName ) + ".exe"
            if File( cEXEName )
               FErase( cEXEName ) // We are forced to delete it as we don't know  
            endif                 // its debugger status
         endif   

         if ( nAt := ::SearchFileInFolder( cFileName, .F. ) ) != 0
            ::oFld:aDialogs[ nAt ]:aControls[ 1 ]:Save()
            if nAt == ::oFld:nOption
               ::oBtnSave:Disable()
            endif   
         endif

         do case
            case Lower( cFileExt( cFileName ) ) == "prg"
                 cTargetFileName = cFilePath( cFileName ) + ;
                                   cFileNoExt( cFileName ) + ".obj"
                 if ! File( cTargetFileName )
                    ::CompilePRG( cFileName, lDebug )
                 else
                    if FDate( cTargetFileName ) <= FDate( cFileName ) .and. ;
                       ( TimeToSec( FTime( cFileName ) ) - ;
                         TimeToSec( FTime( cTargetFileName ) ) ) > 0
                       ::CompilePRG( cFileName, lDebug )
                    endif        
                 endif   
                 
            case Lower( cFileExt( cFileName ) ) == "rc"     
                 cTargetFileName = cFilePath( cFileName ) + ;
                                   cFileNoExt( cFileName ) + ".res"
                 if ! File( cTargetFileName )
                    ::CompileRC( cFileName )
                 else
                    if FDate( cTargetFileName ) <= FDate( cFileName ) .and. ;
                       ( TimeToSec( FTime( cFileName ) ) - ;
                         TimeToSec( FTime( cTargetFileName ) ) ) > 0
                       ::CompileRC( cFileName )
                    endif        
                 endif   
         endcase
      endif             
   next   

   ::Link()

return nil

//----------------------------------------------------------------------------//

METHOD BuildRight() CLASS TFivedit

   local nFunc := 1, oPopup
   
   MENU oPopup POPUP	
      MENUITEM "Copy list to clipboard"
   ENDMENU   		
   
   @ ::oWnd:oBar:nBottom, ::nRightPos + 5 LISTBOX ::oFunList ;
      VAR nFunc ITEMS { "" } ;
      OF ::oWnd SIZE 195, ::oWnd:nHeight PIXEL 

   ::oFunList:bKeyDown = { | nKey | If( nKey == VK_RETURN,;
      Eval( ::oFunList:bLDblClick ),) } 
      
   ::oFunList:SetColor( 0, CLR_GRAY )   

   ::oFunList:bLDblClick = ;
      { || ::oEditor:GotoLine( ::aFunLines[ nFunc ][ 1 ] ),;
                               ::oEditor:SetFocus() }   
          
   ::oFunList:bRClicked = ;
   	{ | nRow, nCol | ::ShowRightPopup( nRow, nCol, ::oFunList ) }        
          			 
return nil                              

//----------------------------------------------------------------------------//

METHOD CopyFuncListToClipboard() CLASS TFivedit

   local cList := "", oClp
   
   AEval( ::oFunList:aItems, { | cItem | cList += cItem + CRLF } )
   
   DEFINE CLIPBOARD oClp OF ::oWnd	

   oClp:SetText( cList )
      		
   oClp:End()

return nil

//----------------------------------------------------------------------------//

METHOD Exit() CLASS TFivedit

   local n
   
   for n = Len( ::oFld:aDialogs ) to 1 step -1
      ::oFld:SetOption( n )
      ::oFld:aDialogs[ n ]:aControls[ 1 ]:Close()
   next
   
   ::oWnd:End()
   
return nil      

//----------------------------------------------------------------------------//

METHOD FillFuncList() CLASS TFivedit

   local nLines := ::oEditor:GetLineCount(), n 
   local aFuncs := {}, aTemp := {}, cToken, cLine
    
   ::aFunLines = {}

   for n = 1 to nLines
      cToken = Lower( Left( cLine := LTrim( ::oEditor:GetLine( n ) ), 4 ) )
      cLine = AllTrim( StrTran( cLine, CRLF, "" ) )
      if cToken $ "static func,func,proc,clas,meth,hb_f"
         AAdd( aTemp, { cLine, n, n, Upper( cLine ) } )
      endif
      if cToken $ "retu" .or. Left( cToken, 1 ) == "}"
         if ! Empty( ATail( aTemp ) )
            ATail( aTemp )[ 3 ] = n
         endif   
      endif
   next

   if ! ::lDuplicates
      while ( n := AScan( aTemp, { | a | "METHOD" $ a[ 4 ] .and. ;
                                       ! "CLASS" $ a[ 4 ] .and. ;
                                       ! "INLINE" $ a[ 4 ] .and. ;
                                       ! "VIRTUAL" $ a[ 4 ] } ) ) != 0
    	   ADel( aTemp, n )
      	ASize( aTemp, Len( aTemp ) - 1 )
      end
   endif      	                        

   ASort( aTemp,,, { | x, y | x[ 4 ] < y[ 4 ] } )
   for n = 1 to Len( aTemp )
     AAdd( ::aFunLines, { aTemp[ n ][ 2 ], aTemp[ n ][ 3 ] } )
     AAdd( aFuncs, aTemp[ n ][ 1 ] )
   next     

   ::oFunList:SetItems( aFuncs )
   
return nil         

//----------------------------------------------------------------------------//

METHOD NewFile() CLASS TFivedit

   local cFileName := "noname1.prg", n := 2
   
   while File( cFileName )
      cFileName = "noname" + AllTrim( Str( n++ ) ) + ".prg"
   end   
   
   if ! File( cFileName )
      MemoWrit( cFileName, '#include "FiveWin.ch"' + CRLF + CRLF + ;
                           'function Main()' + CRLF + CRLF + ;
                           '   MsgInfo( "Hello world!" )' + CRLF + CRLF + ;
                           'return nil' )
   endif
   
   ::OpenFile( cFileName )
   
return nil      

//----------------------------------------------------------------------------//

METHOD OpenFile( cFileName ) CLASS TFivedit

   local oItem
   
   DEFAULT cFileName := cGetFile( FWString( "Program file" ) + ;
                                  " (*.prg) |*.prg|" + ;
                                  FWString( "Header file" ) + ;
                                  " (*.ch) |*.ch|" + ;
                                  FWString( "Resource file" ) + ;
                                  " (*.rc) |*.rc|" + ;
                                  FWString( "Open any file" ) + ;
                                  " (*.*) |*.*|",;
                                  FWString( "Select a file to open" ) )

   if Empty( cFileName )
      return nil
   endif   

   if ! File( cFileName )
      MsgAlert( cFileName + " " + FWString( "does not exist" ) )
      return nil
   endif 
   
   ::BuildEditor( cFileName )
   ::oEditor:OpenFile( cFileName )
   ::oEditor:SetFocus() 

   ::oMruPrgs:Save( cFileName )  
        
return nil

//----------------------------------------------------------------------------//

METHOD OpenProject( cPrjName ) CLASS TFivedit

   local hProject, cFileName, cPreference
   
   DEFAULT cPrjName := cGetFile( "*.prj", "Select a Project file" )

   if ! Empty( ::oTree:Cargo ) .and. ::oTree:Cargo != cPrjName
      ::SaveProject()
      ::CloseProject()
   endif   
   
   if cPrjName == ::oTree:Cargo
      MsgAlert( FWString( "Please close the current project" ) + ;
                CRLF + FWString( "before opening it again" ) )
      return nil
   endif   
   
   if Len( ::oTree:aItems ) > 0
      ::CloseProject()
   endif   
   
   if ! Empty( cPrjName ) .and. File( cPrjName )
      ::oTree:aItems[ 1 ]:SetText( cPrjName )
      hProject = hb_iniRead( cPrjName )

      for each cFileName in hProject[ "files" ]
         ::oTree:aItems[ 1 ]:Add( cFileName, 1 )
      next   

      ::oTree:Expand()
      ::oTree:Cargo = cPrjName
      ::oMruPrjs:Save( cPrjName )

      ::LoadPreferences( cPrjName )
   endif
   
return nil         

//----------------------------------------------------------------------------//

METHOD Parameters() CLASS TFivedit

   local oDlg 
   
   DEFAULT ::cParameters := Space( 200 )
   
   DEFINE DIALOG oDlg RESOURCE "Parameters" ;
      TITLE FWString( "Command line parameters" )
   
   REDEFINE GET ::cParameters ID 100 OF oDlg
   
   ACTIVATE DIALOG oDlg CENTERED

return nil

//----------------------------------------------------------------------------//

METHOD Preferences() CLASS TFivedit

   local oDlg, oFolder, oThis := Self
   local oGetTextColor, cTextColor := Space( 100 )
   local oGetBackColor, cBackColor := Space( 100 )
   local oGetCaretBackColor, cCaretBackColor := Space( 100 )
   
   DEFINE DIALOG oDlg RESOURCE "preferences" ;
      TITLE FWString( "Preferences" )
   
   REDEFINE FOLDER oFolder ID 100 OF oDlg ;
      PROMPTS "&FWH", "&Harbour", FWString( "&C compiler" ),;
              FWString( "source editor" ), FWString( "&general" ) ;
      DIALOGS "fwh", "Harbour", "CCompiler", "editor", "general"

   REDEFINE GET ::cFWHPath ID 110 OF oFolder:aDialogs[ 1 ]
   
   REDEFINE COMBOBOX ::cFWHMode ;
      ITEMS { "32 bits", "64 bits" } ;
      ID 120 OF oFolder:aDialogs[ 1 ] ;
      ON CHANGE MsgInfo( FWString( "Not implemented yet" ) )

   REDEFINE GET ::cFWHExtraLibs ID 130 OF oFolder:aDialogs[ 1 ]
   
   REDEFINE GET ::cHarbPath  ID 110 OF oFolder:aDialogs[ 2 ]

   REDEFINE COMBOBOX ::cHarbComp ITEMS { "Harbour", "xHarbour" } ;
      ID 115 OF oFolder:aDialogs[ 2 ]

   REDEFINE GET ::cHarbFlags ID 120 OF oFolder:aDialogs[ 2 ]

   REDEFINE GET ::cHarbExtraLibs ID 130 OF oFolder:aDialogs[ 2 ]

   REDEFINE GET ::cCCompilerPath  ID 110 OF oFolder:aDialogs[ 3 ]

   REDEFINE GET ::cCCompilerFlags ID 120 OF oFolder:aDialogs[ 3 ]

   REDEFINE GET ::cCCompilerExtraLibs ID 130 OF oFolder:aDialogs[ 3 ]
   
   REDEFINE GET oGetTextColor VAR cTextColor ID 110 OF oFolder:aDialogs[ 4 ] ;
      ACTION oGetTextColor:SetColor( 0, ChooseColor( oGetTextColor:nClrPane ) ) ;
      COLOR 0, ::nTextColor

   REDEFINE GET oGetBackColor VAR cBackColor ID 120 OF oFolder:aDialogs[ 4 ] ;
      ACTION oGetBackColor:SetColor( 0, ChooseColor( oGetBackColor:nClrPane ) ) ;
      COLOR 0, ::nBackColor

   REDEFINE GET oGetCaretBackColor VAR cCaretBackColor ID 130 ;
      OF oFolder:aDialogs[ 4 ] ;
      ACTION oGetCaretBackColor:SetColor( 0, ChooseColor( oGetCaretBackColor:nClrPane ) ) ;
      COLOR 0, ::nCaretBackColor
    
   REDEFINE GET ::cFontInfo ID 140 OF oFolder:aDialogs[ 4 ] ;
      ACTION ( oThis:oEditor:SelFont(),;
               oThis:cFontInfo := FontToText( oThis:oEditor:oFont ) )
   
   REDEFINE COMBOBOX ::cTheme ;
      ITEMS { "Standard", "Dark", "Norton" } ;
      ID 150 OF oFolder:aDialogs[ 4 ] ;
      ON CHANGE If( oThis:cTheme == "Dark",;
      		        oThis:SetDarkTheme( oGetTextColor, oGetBackColor, oGetCaretBackColor ),)		
   	
   REDEFINE COMBOBOX ::nLanguage ;
      ITEMS { "English", "Spanish", "French",;
              "Portugese", "German", "Italian" } ;
      ID 100 OF oFolder:aDialogs[ 5 ] ;
      ON CHANGE ( FWSetLanguage( oThis:nLanguage ),;
                  oThis:oWnd:oBar:End(), oThis:BuildButtonBar(),;
                  oThis:oWnd:oMenu:End(),;
                  oThis:oWnd:SetMenu( oThis:BuildMenu() ),;
                  If( Empty( oThis:oTree:Cargo ),;
                  oThis:oTree:aItems[ 1 ]:SetText( FWString( "Project" ) ),),;
                  oThis:oWnd:oMsgBar:End(), oThis:BuildMsgBar() )

   REDEFINE CHECKBOX ::lDuplicates ID 110 OF oFolder:aDialogs[ 5 ] ;
      ON CHANGE ::FillFuncList()
   
   REDEFINE BUTTON ID 1 OF oDlg ;
      ACTION ( ::nTextColor := oGetTextColor:nClrPane,;
               ::nBackColor := oGetBackColor:nClrPane,;
               ::oEditor:SetColor( ::nTextColor, ::nBackColor ),;
               ::nCaretBackColor := oGetCaretBackColor:nClrPane,;
               ::oEditor:SetCaretLineBack( ::nCaretBackColor ),;
               ::SavePreferences(), oDlg:End() )
   
   ACTIVATE DIALOG oDlg CENTERED

return nil

//----------------------------------------------------------------------------//

METHOD ProjectAddItem() CLASS TFivedit

   local oItem
   
   if ! ::SearchFileInTree( ::oFld:aPrompts[ ::oFld:nOption ] )
      oItem = ::oTree:aItems[ 1 ]:Add( ::oEditor:cFileName, 1 )
      ::oTree:Expand() 
      ::oTree:Select( oItem )
   endif   
   
return nil   

//----------------------------------------------------------------------------//

METHOD ProjectDelItem() CLASS TFivedit

   local nAt, cFileName

   if ::oTree:GetSelected() == ::oTree:aItems[ 1 ]
      return nil
   endif
      
   cFileName = ::oTree:GetSelected():cPrompt   
   ::oTree:GetSelected():End()

return nil

//----------------------------------------------------------------------------//

METHOD ProjectOpenFile() CLASS TFivedit

   local oItem := ::oTree:GetSelected()
   
   if oItem == ::oTree:aItems[ 1 ]
      return nil
   endif
   
   if ::SearchFileInFolder( oItem:cPrompt ) == 0
      if File( oItem:cPrompt )
         ::OpenFile( oItem:cPrompt )
      else   
         MsgAlert( FWString( "File" ) + " " + oItem:cPrompt + " " + ;
                   FWString( "not found" ) )
      endif
   endif      

return nil

//----------------------------------------------------------------------------//

METHOD RunScript() CLASS TFivedit
  
   local oHrb, cResult, bOldError

   FReOpen_Stderr( "comp.log", "w" )
   
   oHrb = HB_CompileFromBuf( StrTran( ::oEditor:GetText(), "Main", "__Main" ),;
                             "-n", "-I" + AllTrim( ::cFWHPath ) + "\include",;
                             "-I" + AllTrim( ::cHarbPath ) + "\include" )

   if "Error" $ MemoRead( "comp.log" )
      MsgInfo( MemoRead( "comp.log" ), "Error" )
   else   
      if ! Empty( oHrb )
         BEGIN SEQUENCE
         bOldError = ErrorBlock( { | o | DoBreak( o ) } )
         hb_HrbRun( oHrb )
         END SEQUENCE
         ErrorBlock( bOldError )
      endif
   endif   

return nil

//----------------------------------------------------------------------------//

METHOD SelectFunction() CLASS TFivedit

   local nAt, nLine := ::oEditor:nLine()
   
   if ( nAt := AScan( ::aFunLines, {| u | u[ 1 ] <= nLine .and. ;
      u[ 2 ] >= nLine } ) ) != 0
      ::oFunList:Select( nAt )
   endif
   
return nil      

//----------------------------------------------------------------------------//

METHOD ShowProject() CLASS TFivedit

   ::oVSplitL:SetPosition( PROJECT_WIDTH )
   ::oTree:nWidth := PROJECT_WIDTH - 1
   ::oTree:Show() 
                           									
return nil

//----------------------------------------------------------------------------//

METHOD SetDarkTheme( oGetTextColor, oGetBackColor, oGetCaretBackColor ) CLASS TFivedit

   local n

   AEval( ::oBar:aControls, { | o | o:nClrText := CLR_WHITE } )
   ::oBar:bClrGrad = { | lInvert | If( ! lInvert,;
                         { { 0.25, RGB( 40, 40, 43 ), RGB( 40, 40, 43 ) },;
                           { 0.75, RGB( 20, 20, 20 ), RGB( 20, 20, 20 ) } },;
                         { { 0.25, RGB( 255, 253, 222 ), RGB( 255, 231, 151 ) }, ;
                           { 0.75, RGB( 255, 215,  84 ), RGB( 255, 233, 162 ) } } ) }
   ::oBar:Refresh()									

   oGetTextColor:nClrPane = CLR_WHITE
   oGetBackColor:nClrPane = nRgb( 40, 40, 43 )
   oGetCaretBackColor:nClrPane = 0	
   	
   for n = 1 to Len( ::oFld:aDialogs )
      ::oFld:aDialogs[ n ]:aControls[ 1 ]:SetColor( CLR_WHITE, nRgb( 40, 40, 43 ) )
   next   		

return nil

//----------------------------------------------------------------------------//

METHOD ShowFuncList() CLASS TFivedit

   ::oVSplitR:SetPosition( ::oWnd:nWidth - 200 )
   ::oFunList:nWidth := 199
   ::oFunList:Show()

return nil

//----------------------------------------------------------------------------//

METHOD ShowRightPopup( nRow, nCol, oWnd ) CLASS TFivedit

   local oPopup
   
   MENU oPopup POPUP
      MENUITEM FWString( "Copy to clipboard" ) RESOURCE "copy16" ;
         ACTION ::CopyFuncListToClipboard()		
   ENDMENU
   
   ACTIVATE MENU oPopup AT nRow, nCol OF oWnd
   
return nil   	
 
//----------------------------------------------------------------------------//

METHOD ShowResult() CLASS TFivedit

   ::oHSplitC:SetPosition( ::oWnd:nHeight - 200 )
   ::oMemo:Show()
   
return nil   	

//----------------------------------------------------------------------------//

METHOD Tools() CLASS TFivedit

   local oDlg, oLbx, aTools := {}, aNames := {}
   local cItem, cName, cParams, cPrompt
   local oName, oParams, oPrompt
   local cTool, n := 1, oIni
   local cIniFile := "c:\users\" + WNetGetUser() + ;
    	               "\local\appdata\fivedit\fivedit.ini"	
   
   cItem := cName := cParams := cPrompt := Space( 200 )

   if File( cIniFile )
      INI oIni FILENAME cIniFile
         GET cTool SECTION "Tools" ENTRY Str( n++ ) OF oIni

         if Empty( cTool )
            AAdd( aNames, "&Cmd Window,cmd.exe" )
            AAdd( aTools, "&Cmd Window" )
            AAdd( aNames, "&FiveDBU,fivedbu.exe" ) 
            AAdd( aTools, "&FiveDBU" )
         endif   

         while ! Empty( cTool )
            AAdd( aNames, cTool )
            AAdd( aTools, StrToken( cTool, 1, "," ) )

            GET cTool SECTION "Tools" ;
               ENTRY Str( n, If( n < 10, 1, 2 ) ) OF oIni

            n++
         end
      ENDINI      
   endif   
   
   DEFINE DIALOG oDlg RESOURCE "Tools" ;
      TITLE FWString( "Tools settings" )
   
   REDEFINE LISTBOX oLbx VAR cItem ID 110 OF oDlg ITEMS aTools ;
      ON CHANGE cItem := aNames[ oLbx:GetPos() ],;
                cName := PadR( StrToken( cItem, 2, "," ), 50 ),;
                oName:Refresh(),;
                cParams := PadR( StrToken( cItem, 3, "," ), 20 ),;
                oParams:Refresh(),;
                cPrompt := PadR( StrToken( cItem, 1, "," ), 20 ),;
                oPrompt:Refresh()
   
   REDEFINE BUTTON ID 120 OF oDlg ACTION oLbx:SwapUp()
   
   REDEFINE BUTTON ID 130 OF oDlg ACTION oLbx:SwapDown()
   
   REDEFINE BUTTON ID 140 OF oDlg ;
      ACTION AddTool( oLbx, aNames )

   REDEFINE BUTTON ID 150 OF oDlg ;
      ACTION ADel( aNames, oLbx:GetPos() ),;
             ASize( aNames, Len( aNames ) - 1 ),;
             oLbx:Del( oLbx:GetPos() ),;
             Eval( oLbx:bChange )
                   
   REDEFINE GET oName VAR cName ID 160 OF oDlg 
   
   REDEFINE GET oParams VAR cParams ID 170 OF oDlg
   
   REDEFINE GET oPrompt VAR cPrompt ID 180 OF oDlg ;
      ON CHANGE ( cPrompt := oPrompt:GetText(),;
                  aNames[ oLbx:GetPos() ] := AllTrim( cPrompt ) + "," + ;
                  AllTrim( cName ),;
                  oLbx:Modify( AllTrim( cPrompt ), oLbx:GetPos() ), .T. )
   
   REDEFINE BUTTON ID ID_OK OF oDlg ;
      ACTION SaveTools( aNames ), ::MenuSetTools(), oDlg:End()   
   
   ACTIVATE DIALOG oDlg CENTERED ;
      ON INIT ( If( Len( aTools ) > 0, Eval( oLbx:bChange ),), .T. )
   
return nil   

//----------------------------------------------------------------------------//

static function AddTool( oLbx, aNames )

   local cApp := Lower( cGetFile( FWString( "Executable" ) + ;
                                  " (*.exe) |*.exe|" + ;
                                  FWString( "Batch file" ) + ;
                                  " (*.bat) |*.bat|",;
                                  FWString( "Select a tool" ) )	)										 
   local cPrompt

   if ! Empty( cApp ) .and. File( cApp )
      cPrompt = StrCapFirst( cFileNoPath( cFileNoExt( cApp ) ) )
      oLbx:Add( cPrompt )
      AAdd( aNames, cPrompt + "," + cApp )
      oLbx:GoBottom()
   endif

return nil

//----------------------------------------------------------------------------//

static function SaveTools( aNames )

   local oIni, n

   INI oIni FILENAME "c:\users\" + WNetGetUser() + ;
                   	"\appdata\local\fivedit\fivedit.ini"
      for n = 1 to Len( aNames )
         SET SECTION "Tools" ENTRY Str( n, If( n < 10, 1, 2 ) ) ;
            TO aNames[ n ] OF oIni
      next
      for n = Len( aNames ) + 1 to 10
         SET SECTION "Tools" ENTRY Str( n, If( n < 10, 1, 2 ) ) ;
            TO "" OF oIni
      next
   ENDINI

return nil

//----------------------------------------------------------------------------//

METHOD HideFuncList() CLASS TFivedit

   ::oFunList:Hide()
   ::oVSplitR:SetPosition( ::oWnd:nWidth - 15 )
   ::oVSplitR:Adjust()
   	
   AEval( ::oFld:aDialogs, { | o | o:aControls[ 1 ]:AdjClient() } )   

return nil

//----------------------------------------------------------------------------//

METHOD HideProject() CLASS TFivedit

   ::oTree:Hide()
   ::oVSplitL:SetPosition( 0 )
   ::oVSplitL:Adjust()
   	
return nil

//----------------------------------------------------------------------------//

METHOD HideResult() CLASS TFivedit

   ::oMemo:Hide()
   ::oHSplitC:SetPosition( ::oWnd:nHeight - ::oBar:nHeight - ;
                           ::oMsgBar:nHeight() )
   
return nil   	

//----------------------------------------------------------------------------//

METHOD Link() CLASS TFivedit

   local n, cBuildBat, cFileName, cOBJs := "", cRESs := ""
   local cPrjPath := Lower( cFilePath( ::oTree:Cargo ) )
   local cEXEName
   
   for n = 1 to Len( ::oTree:aItems[ 1 ]:aItems )
      cFileName = ::oTree:aItems[ 1 ]:aItems[ n ]:cPrompt   
      
      if n == 1
         cEXEName = cFileNoExt( cFileName )
      endif   
      
      do case
         case Lower( cFileExt( cFileName ) ) == "prg"
            cOBJs += Lower( cFilePath( cFileName ) ) + ;
                     cFileNoExt( cFileName ) + ".obj" + " "

         case Lower( cFileExt( cFileName ) ) == "res"
            cRESs += Lower( cFileName ) + " "

      endcase               
   next

   if File( cPrjPath + cEXEName + ".exe" )
      WaitRun( cPrjPath + cEXEName + ".exe", 0 )
      return nil
   endif    

   cBuildBat = AllTrim( ::cCCompilerPath ) + "\bin\ilink32 " + ;
               ::cCCompilerFlags + " " + AllTrim( ::cCCompilerPath ) + ;
               "\lib\c0w32.obj " + cOBJs + ", " + ;
               cPrjPath + cEXEName + ".exe, " + ;
               cPrjPath + cEXEName + ".map, " + ;
               ::AddHarbourLibs() + ;
               ::AddBorlandLibs() + ;
               ", " + cRESs + " > " + cPrjPath + "build.log" + CRLF 

   MemoWrit( cPrjPath + "fivedit.bat", cBuildBat )
   WaitRun( cPrjPath + "fivedit.bat", 0 )
   
   ::oMemo:SetText( ::oMemo:GetText() + ;
                    MemoRead( cPrjPath + "build.log" ) )
   ::oMemo:GoBottom()

   if File( cPrjPath + cEXEName + ".exe" )
      WaitRun( cPrjPath + cEXEName + ".exe" )
   endif    
   
return nil

//----------------------------------------------------------------------------//

METHOD LoadPanelsSettings() CLASS TFivedit

   local oIni, lProject, lFuncList, lResult

   lMkDir( "c:\users\" + WNetGetUser() + "\appdata\local\fivedit" )

   INI oIni FILENAME "c:\users\" + WNetGetUser() + ;
                   	"\appdata\local\fivedit\fivedit.ini"
      GET lProject SECTION "panels" ENTRY "1" ;
         DEFAULT .T. OF oIni

      GET lFunclist SECTION "panels" ENTRY "2" ;
         DEFAULT .T. OF oIni

      GET lResult SECTION "panels"  ENTRY "3" ;
         DEFAULT .T. OF oIni
   ENDINI
   
   if ! lProject
      ::HideProject()
      ::oItemProject:SetCheck( .F. )       	
   endif

   if ! lFuncList
      ::HideFuncList()
      ::oItemFuncList:SetCheck( .F. )       	
   endif

   if ! lResult
      ::HideResult()
      ::oItemResult:SetCheck( .F. )       	
   endif

return nil

//----------------------------------------------------------------------------//

METHOD LoadPreferences( cIniFile ) CLASS TFivedit

   local oIni
   local cFWHPath, cFWHMode, cFWHExtraLibs
   local cHarbPath, cHarbComp, cHarbFlags, cHarbExtraLibs
   local cCCompilerPath, cCCompilerFlags, cCCompilerExtraLibs

   DEFAULT cIniFile := "c:\users\" + WNetGetUser() + ;
       	  "\appdata\local\fivedit\fivedit.ini"

   INI oIni FILENAME cIniFile
      GET cFWHPath SECTION "preferences" ENTRY "1" ;
         DEFAULT ::cFWHPath OF oIni

      GET cFWHMode SECTION "preferences" ENTRY "2" ;
         DEFAULT ::cFWHMode OF oIni

      GET cFWHExtraLibs SECTION "preferences" ENTRY "3" ;
         DEFAULT ::cFWHExtraLibs OF oIni
         
      GET cHarbPath SECTION "preferences" ENTRY "4" ;
         DEFAULT ::cHarbPath OF oIni

      GET cHarbComp SECTION "preferences" ENTRY "5" ;
         DEFAULT ::cHarbComp OF oIni
         
      GET cHarbFlags SECTION "preferences" ENTRY "6" ;
         DEFAULT ::cHarbFlags OF oIni
         
      GET cHarbExtraLibs SECTION "preferences" ENTRY "7" ;
         DEFAULT ::cHarbExtraLibs OF oIni
         
      GET cCCompilerPath SECTION "preferences" ENTRY "8" ;
         DEFAULT ::cCCompilerPath OF oIni
         
      GET cCCompilerFlags SECTION "preferences" ENTRY "9" ;
         DEFAULT ::cCCompilerFlags OF oIni
         
      GET cCCompilerExtraLibs SECTION "preferences" ENTRY "10" ;
         DEFAULT ::cCCompilerExtraLibs OF oIni

      GET ::nTextColor SECTION "preferences" ENTRY "11" ;
         DEFAULT ::nTextColor OF oIni

      GET ::nBackColor SECTION "preferences" ENTRY "12" ;
         DEFAULT ::nBackColor OF oIni

      GET ::nCaretBackColor SECTION "preferences" ENTRY "13" ;
         DEFAULT ::nCaretBackColor OF oIni

      GET ::cFontInfo SECTION "preferences" ENTRY "14" ;
         DEFAULT ::cFontInfo OF oIni

      GET ::nLanguage SECTION "preferences" ENTRY "15" ;
         DEFAULT ::nLanguage OF oIni

      GET ::cTheme SECTION "preferences" ENTRY "16" ;
         DEFAULT ::cTheme OF oIni

      GET ::lDuplicates SECTION "preferences" ENTRY "17" ;
         DEFAULT ::lDuplicates OF oIni

      GET ::cDirName SECTION "preferences" ENTRY "18" ;
         DEFAULT ::cDirName OF oIni

   ENDINI

   ::cFWHPath   = PadR( cFWHPath, Len( ::cFWHPath ) )
   ::cFWHMode   = PadR( cFWHMode, Len( ::cFWHMode ) )
   ::cFWHExtraLibs = PadR( cFWHExtraLibs, Len( ::cFWHExtraLibs ) )
   
   ::cHarbPath  = PadR( cHarbPath, Len( ::cHarbPath ) )
   ::cHarbComp  = PadR( cHarbComp, Len( ::cHarbComp ) )
   ::cHarbFlags = PadR( cHarbFlags, Len( ::cHarbFlags ) )
   ::cHarbExtraLibs = PadR( cHarbExtraLibs, Len( ::cHarbExtraLibs ) )
   
   ::cCCompilerPath = PadR( cCCompilerPath, Len( ::cCCompilerPath ) )
   ::cCCompilerFlags = PadR( cCCompilerFlags, Len( ::cCCompilerFlags ) )
   ::cCCompilerExtraLibs = PadR( cCCompilerExtraLibs,;
                                Len( ::cCCompilerExtraLibs ) )
   FWSetLanguage( ::nLanguage )
   
   ::cFontInfo = PadR( ::cFontInfo, 200 )

   do case
      case ::cTheme == "Dark"
           ::SetDarkTheme()
   endcase        			  

   if ! Empty( ::cDirName )
      SetCurDir( ::cDirName )
   endif

return nil

//----------------------------------------------------------------------------//

METHOD MenuSetTools() CLASS TFivedit

   local oIni, n := 1, cTool, oItem
   local cIniFile := "c:\users\" + WNetGetUser() + ;
                     "\appdata\local\fivedit\fivedit.ini"

   if File( cIniFile )
      INI oIni FILENAME cIniFile
         GET cTool SECTION "Tools" ENTRY Str( n++ ) OF oIni

         if ! Empty( cTool )
            ::oMenuTools:DelItems()
            ::oMenuTools:l2007 = .T.
            SysRefresh()

            while ! Empty( cTool )
               ::oMenuTools:Add( oItem := ;
                  TMenuItem():New( StrToken( cTool, 1, "," ),;
                                   StrToken( cTool, 2, "," ) ) )
               
               oItem:bAction = &( '{ || WinExec( "' + ;
                                  StrToken( cTool, 2, "," ) + '" ) }' )            

               GET cTool SECTION "Tools" ;
                  ENTRY Str( n, If( n < 10, 1, 2 ) ) OF oIni

               n++
            end

            ::oMenuTools:Add( TMenuItem():New() )

            ::oMenuTools:Add( TMenuItem():New( FWString( "Settings" ) + "...",;
                                               FWString( "Configure this menu" ),,,;
                                               { || ::Tools() } ) )
         else   
            MENUITEM FWString( "&Cmd Window" ) ;
               RESOURCE "cmd16" ACTION WinExec( "cmd.exe" )    
         
            MENUITEM "&FiveDBU" ;
               RESOURCE "fivetech16" ;
               ACTION WinExec( "fivedbu.exe" )    
               
            SEPARATOR   

            MENUITEM FWString( "Settings" ) + "..." ;
               MESSAGE FWString( "Configure this menu" ) ;
               ACTION ::Tools()
         endif      
      ENDINI
   else
      MENUITEM FWString( "&Cmd Window" ) ;
         RESOURCE "cmd16" ACTION WinExec( "cmd.exe" )    
         
      MENUITEM FWString( "&FiveDBU" ) ;
         RESOURCE "fivetech16" ;
         ACTION WinExec( "fivedbu.exe" )
         
      SEPARATOR       
      
      MENUITEM FWString( "Settings" ) + "..." ;
         MESSAGE FWString( "Configure this menu" ) ;
         ACTION ::Tools()
   endif   

return nil

//----------------------------------------------------------------------------//

static function DoBreak( oError )

   local cInfo := oError:operation, n

   if ValType( oError:Args ) == "A"
      cInfo += "   Args:" + CRLF
      for n = 1 to Len( oError:Args )
         MsgInfo( oError:Args[ n ] )
         cInfo += "[" + Str( n, 4 ) + "] = " + ValType( oError:Args[ n ] ) + ;
                   "   " + cValToChar( oError:Args[ n ] ) + CRLF
      next
   endif

   MsgStop( oError:Description + CRLF + cInfo,;
            FWString( "Script error at line" ) + ": " + ;
            AllTrim( Str( ProcLine( 2 ) ) ) )

   BREAK

return nil

//----------------------------------------------------------------//

METHOD CheckBorlandSetupFiles() CLASS TFivedit

   if ! File( AllTrim( ::cCCompilerPath ) + "\bin\bcc32.cfg" )
      MemoWrit( AllTrim( ::cCCompilerPath ) + "\bin\bcc32.cfg",;
                '-I"' + AllTrim( ::cCCompilerPath ) + "\include;" + ;
                AllTrim( ::cCCompilerPath ) + "\include\dinkumware" + ;
                '"' + CRLF + ;
                '-L"' + AllTrim( ::cCCompilerPath ) + "\lib;" + ;
                AllTrim( ::cCCompilerPath ) + "\lib\psdk" + '"' + CRLF )
   endif             

   if ! File( AllTrim( ::cCCompilerPath ) + "\bin\ilink32.cfg" )
      MemoWrit( AllTrim( ::cCCompilerPath ) + "\bin\ilink32.cfg",;
                '-L"' + AllTrim( ::cCCompilerPath ) + "\lib;" + ;
                AllTrim( ::cCCompilerPath ) + "\lib\psdk" + ;
                '"' + CRLF )
   endif             

return nil

//----------------------------------------------------------------//

METHOD Close() CLASS TFivedit

   ::oFld:DelItemPos( ::oFld:nOption )

return nil

//----------------------------------------------------------------------------//

METHOD CloseProject() CLASS TFivedit

   ::oTree:aItems[ 1 ]:DeleteBranches()
   ::oTree:Cargo = nil
   ::oTree:aItems[ 1 ]:SetText( "Project" )
                     
return nil                     

//----------------------------------------------------------------------------//

METHOD CompilePRG( cFileName, lDebug ) CLASS TFivedit

   local cBuildBat, cPrjPath := Lower( cFilePath( ::oTree:Cargo ) )

   cBuildBat = AllTrim( ::cHarbPath ) + "\bin\harbour %1 " + ;
               "/i" + AllTrim( ::cHarbPath ) + "\include;" + ;
               AllTrim( ::cFWHPath ) + "\include " + ;
               AllTrim( ::cHarbFlags ) + If( lDebug, " /b", "" ) + ;
               " /o" + Lower( cFilePath( cFileName ) ) + ;
               cFileNoExt( cFileName ) + ".c" + ;
               " > comp1.log 2> comp2.log" + CRLF + ;
               "if errorlevel 0 " + AllTrim( ::cCCompilerPath ) + "\bin\bcc32 -c" + ;
               " -I" + AllTrim( ::cHarbPath ) + "\include" + ;
               " -o" + Lower( cFilePath( cFileName ) ) + ;
               cFileNoExt( cFileName ) + ;
               ".obj" + " %1.c" + CRLF

   // FReOpen_Stderr( "build.log", "w" )
   
   MemoWrit( cPrjPath + "fivedit.bat", cBuildBat )
   
   WaitRun( cPrjPath + "fivedit.bat " + ;
            Lower( cFilePath( cFileName ) ) + ;
            cFileNoExt( cFileName ) + If( lDebug, " /b", "" ), 0 )
   
   ::oMemo:SetText( ::oMemo:GetText() + ;
                    MemoRead( cPrjPath + "comp1.log" ) + CRLF + ;
                    MemoRead( cPrjPath + "comp2.log" ) + CRLF )
   ::oMemo:GoBottom()

return nil

//----------------------------------------------------------------------------//

METHOD CompileRC( cFileName ) CLASS TFivedit

   local cBuildBat, cPrjPath := Lower( cFilePath( ::oTree:Cargo ) )

   cBuildBat = AllTrim( ::cCCompilerPath ) + "\bin\brc32 %1" + ;
               " -i" + AllTrim( ::cCCompilerPath ) + "\include" + ;
               " -fo" + Lower( cFilePath( cFileName ) ) + ;
               cFileNoExt( cFileName ) + ".res" + ;
               " -r -d__FLAT__ " + CRLF

   // FReOpen_Stderr( "build.log", "w" )
   
   MemoWrit( cPrjPath + "fivedit.bat", cBuildBat )
   
   WaitRun( cPrjPath + "fivedit.bat " + Lower( cFilePath( cFileName ) ) + ;
            cFileNoExt( cFileName ) + " > " + cPrjPath + "build.log", 0 )
   
   ::oMemo:SetText( ::oMemo:GetText() + MemoRead( cPrjPath + "build.log" ) )
   ::oMemo:GoBottom()

return nil

//----------------------------------------------------------------------------//

METHOD Run( lDebug ) CLASS TFivedit

   local cBuildBat, cBaseName

   DEFAULT lDebug := .F.

   if Len( ::oTree:aItems[ 1 ]:aItems ) > 0
      return ::BuildProject( lDebug )
   endif   

   if ! File( AllTrim( ::cFWHPath ) + "\include\FiveWin.ch" ) 
      MsgAlert( "Please set the right FWH path from preferences" )
      return nil
   endif   

   if ! File( AllTrim( ::cHarbPath ) + "\bin\harbour.exe" ) 
      MsgAlert( "Please set the right Harbour path from preferences" )
      return nil
   endif   

   if ! File( AllTrim( ::cCCompilerPath ) + "\bin\bcc32.exe" ) 
      MsgAlert( "Please set the right Borland C compiler path from preferences" )
      return nil
   else   
      ::CheckBorlandSetupFiles()
   endif   

   if ! ::oItemResult:lChecked
      ::oItemResult:SetCheck( .T. )	
      ::ShowResult()
   endif   		

   cBaseName = cFileNoExt( ::oEditor:cFileName )

   cBuildBat = AllTrim( ::cHarbPath ) + "\bin\harbour " + cBaseName + ;
               " /i" + AllTrim( ::cHarbPath ) + "\include;" + ;
               AllTrim( ::cFWHPath ) + "\include " + ;
               AllTrim( ::cHarbFlags ) + If( lDebug, "/b", "" ) + ;
               " > comp1.log 2> comp2.log"	+ CRLF + ;
               "if errorlevel 0 " + AllTrim( ::cCCompilerPath ) + "\bin\bcc32 -c" + ;
               " -I" + AllTrim( ::cHarbPath ) + "\include " + ;
               cBaseName + ".c" + CRLF + ; 
               "if errorlevel 0 " + AllTrim( ::cCCompilerPath ) + "\bin\ilink32 " + ;
               AllTrim( ::cCCompilerFlags ) + " " + ;
               AllTrim( ::cCCompilerPath ) +  "\lib\c0w32.obj " + ;
               cBaseName + ".obj, " + cBaseName + ".exe, " + ;
               cBaseName + ".map, " + ;
               ::AddHarbourLibs() + ;
               ::AddBorlandLibs() + " > link.log" + CRLF 

   if ::oEditor:GetModify()
      ::oEditor:Save()
      ::oBtnSave:Disable()
   endif

   // FReOpen_Stderr( "build.log", "w" )
   
   MemoWrit( "fivedit.bat", cBuildBat )
   
   WaitRun( "fivedit.bat", 0 )
   
   ::oMemo:SetText( MemoRead( "comp1.log" ) + ;
                    MemoRead( "comp2.log" ) + ;
                    MemoRead( "link.log" ) )
   ::oMemo:GoBottom()
   
   if File( cBaseName + ".exe" )
      WaitRun( cBaseName + ".exe" )
   endif     

return nil

//----------------------------------------------------------------------------//

CLASS TFiveditWnd FROM TWindow

   CLASSDATA lRegistered INIT .F.

   METHOD HandleEvent( nMsg, nWParam, nLParam )

ENDCLASS

//----------------------------------------------------------------------------//

METHOD HandleEvent( nMsg, nWParam, nLParam ) CLASS TFiveditWnd

   if nMsg == FM_OPENFILE
      OpenClipboard( ::hWnd )
      oFivedit:OpenFile( GetClpData( 1 ) )
      CloseClipboard()
   endif
   
return ::Super:HandleEvent( nMsg, nWParam, nLParam )

//----------------------------------------------------------------------------// 